/****************************************
 * Gestion de l'interface avec le shell
 * init_pipe() démarre le shell et crée les deux pipes
 * envoieshell() envoie une commande à sh (sans '\n' au bout)
 * recoitshell() reçoit une ligne de la réponse (ou NULL après la dernière)
 * execshell() exécute une commande shell sans attendre la réponse
 * erreur() provoque une erreur générale
 * *************************************/

// pour l'instant, ne fonctionne que sous Unix
// pour Windows, voir la page web:
// https://msdn.microsoft.com/en-us/library/ms682499.aspx
//  "Creating a Child Process with Redirected Input and Output"
// ainsi que:
// https://technet.microsoft.com/en-us/library/bb497014.aspx
//  "Chapter 5: Developing Phase: Infrastructure Services"
//  "This chapter discusses the potential coding differences between UNIX and Microsoft® Windows® operating systems with respect to infrastructure services."

#ifndef _GNU_SOURCE
#define _GNU_SOURCE // nécessaire pour fcloseall()
#endif // _GNU_SOURCE
#include "prehtml.h"

/* fichiers inclus uniquement dans cette partie, pour cette raison
 * ils ne sont pas dans prehtml.h */
#include <signal.h>
#include <wait.h>
#include <string>

#ifdef __CYGWIN__
// rien besoin de changer pour CYGWIN
#endif // CYGWIN

FILE *pin=NULL,*pout=NULL;
static int result_fork;
static string envoyee;

static int vmessage(const char *format, va_list ap){
	if (bits8 != 2)
		return vfprintf(stderr, format, ap);
	int n = strlen(format);
	int n2 = n;
	int i,j;
	for (i=0; i<n; i++)
		if (format[i] & 0x80)
			n2++;
	string format2;
	format2.resize(n2);
	j=0;
	for (i=0; i<n; i++){
		if (format[i] & 0x80){
			format2[j++] = 0xC0 | (((unsigned char)format[i])>>6);
			format2[j++] = 0x80 | (format[i]&0x3F);
		} else
			format2[j++] = format[i];
	}
	return vfprintf(stderr, format2.c_str(), ap);
}

int message(const char *format, ...){ // formate un message dans stderr (passe en UTF-8 si option "-u")
	va_list ap;
  	va_start(ap,format);
	return vmessage(format,ap);
}

void erreur(const char *format, ...){
  va_list ap;
  va_start(ap,format);
  dit_ou();
  message("prehtml: ");
  vmessage(format,ap);
  putc('\n',stderr);
  exit(1);
}

void warning(const char *format, ...){
  va_list ap;
  va_start(ap,format);
  dit_ou();
  message("prehtml: attention: ");
  vmessage(format,ap);
  putc('\n',stderr);
}

static vector<string> listetemp;

void creetemp(const char *nom){ /* signale la création d'un fichier temporaire,
			   * pour permettre le nettoyage si nécessaire */
  listetemp.push_back(nom);
}

void fin_urgente(int numsignal){
  char nom[10];
  switch(numsignal){
   case SIGALRM: strcpy(nom,"SIGALRM");break;
   case SIGHUP: strcpy(nom,"SIGHUP");break;
   case SIGINT: strcpy(nom,"SIGINT");break;
   case SIGQUIT: strcpy(nom,"SIGQUIT");break;
   case SIGABRT: strcpy(nom,"SIGABT");break;
   default: sprintf(nom,"n°%-5d",numsignal);
  }
  if (numsignal)
    warning("Interruption reçue: %s\n"
	  "commande éventuellement en cours de traitement: \n"
	  "   <%s %s>\n"
	  "Nettoyage des fichiers temporaires",
	  nom,(fonction.size()?fonction.c_str():"[aucune]"),chaine.c_str());
  else message("Nettoyage des fichiers temporaires\n\n");
  /* la fonction suivante est une extension GNU !!!
   elle sert à fermer les fichiers avant de les refermer;
   il semble que ça marche quand même si on ne la met pas...*/
  fcloseall();
  for(size_t n=0; n<listetemp.size(); n++) /* on a créé des fichiers temporaires*/
    unlink(listetemp[n].c_str());
  listetemp.clear();
  if (numsignal){
    /* je remets le comportement pas défaut, et je relance le signal */
    signal(numsignal, SIG_DFL);
    raise(numsignal);
  }
}


static int alarmstop=0;
void alarme(int numsignal){
  if (alarmstop){
    message("shell: la commande:\n     \"%s\"\n"
	 " ne répond pas: peut-être ceci vient-il de la balise:\n"
	 "     <%s %s>\n"
	 "Vérifiez les guillemets incomplets, crochets, etc...\n\n",
	 envoyee.c_str(),fonction.c_str(),chaine.c_str());
    fin_urgente(0);
    exit(1);
  }
  else
    warning("shell:  la commande:\n     \"%s\"\n"
	    " semble durer un peu longtemps: est-ce cette balise?\n"
	    "     <%s %s>",
	 envoyee.c_str(),fonction.c_str(),chaine.c_str());
}

void init_pipe(void){
  int fd[4];
  pipe(fd);
  pipe(fd+2);
  result_fork=fork();
  if (result_fork==-1) erreur("Le fork n'a pas marché");
  if (result_fork==0){ // process fils
    dup2(fd[0],0); // stdin
    dup2(fd[3],1); // stdout
    setlinebuf(stdin);
    setlinebuf(stdout);
    //execlp("sh","sh",NULL);
	execlp(shell_command,shell_command,NULL);
    erreur("exec '%s' n'a pas marché", shell_command);
  }
  else{
    /* ... car si on ne les ouvre pas dans le même ordre ici, ça bloque */
    pin=fdopen(fd[2],"rb");
    pout=fdopen(fd[1],"wb");
    setlinebuf(pin);
    setlinebuf(pout);
    // on vérifie que ça fonctionne
    fputs("echo\n",pout);
    fflush(pout);
    while(fgetc(pin)!='\n');
  }
  if (signal (SIGALRM,alarme)==SIG_ERR ||
      signal(SIGHUP,fin_urgente)==SIG_ERR ||
      signal(SIGINT,fin_urgente)==SIG_ERR ||
      signal(SIGQUIT,fin_urgente)==SIG_ERR ||
      signal(SIGABRT,fin_urgente)==SIG_ERR )
    message("prehtml: attention: des interruptions n'ont pas pu être détournées\n");
}

void fin_pipe(void){
  if (!pout) erreur("fin_pipe(): pout déjà fermé");
  fclose(pout);
  if (!pin) erreur("fin_pipe(): pin déjà fermé");
  fclose(pin);
  pout=pin=NULL;
  kill(result_fork,SIGTERM);
  wait(NULL);
}


void envoieshell(string &s){  // envoie une commande à sh
  if (shell)
    message("prehtml: shell: > \"%s\"\n",s.c_str());
  char s2[80+1];
  if (!pout) erreur("envoieshell: pipe out non ouvert!");
  envoyee=s;
  fprintf(pout,"echo COMMANDE_TEST;%s\necho %s\n",s.c_str(),CODE_FIN_SHELL);
  fflush(pout);
  alarmstop=1; /* si timeout ici, c'est la fin: le shell ne se lance pas */
  alarm(3); // alarme dans 3 secondes !
  fgets(s2,80,pin);
  alarm(30); // OK, j'ai reçu quelques chose
  alarmstop=0; /* si timeout ici, c'est peut-être que la commande shell est
		* simplement très longue: simple warning */
  if (strcmp(s2,"COMMANDE_TEST\n"))
    erreur("shell: Avant même d'envoyer \"%s\",\n"
	   "  je reçois \"%s\". Erreur interne ?",
	   s.c_str(),s2);
}


int recoitshell(string &s, int retireEOL){ // récupère la réponse de sh et met une ligne
                            // dans s; renvoie 0 si c'est fini, 1 sinon
  s.clear();
  char lignerecue[80+1];


  static int fin_ligne_otee=0;
  int fin_prec;
  int i;

  if (retireEOL) fin_ligne_otee=0;
  if (fin_ligne_otee)
    s="\n";
  fin_ligne_otee=0;
  fin_prec=s.size();
  while(!feof(pin)){
    fgets(lignerecue,80,pin);
    i=strlen(lignerecue);
    if (i) if (lignerecue[i-1]=='\n') {lignerecue[i-1]='\0';fin_ligne_otee=1;}
    s += lignerecue;
    if (fin_ligne_otee) break;
  }
  if (!s.compare(fin_prec,strlen(CODE_FIN_SHELL),CODE_FIN_SHELL)) {
    fin_ligne_otee=0; // le dernier \n doit bien être ôté
    alarm(0); // ma commande shell est finie, donc le timeout aussi
    return 0;
  }
  if (shell)
    message("prehtml: shell: < \"%s\"\n",s.c_str());
  return 1;
}

void execshell(string &s){ // envoie une commande à sh mais néglige la réponse
  string bidon;
  envoieshell(s);
  while (recoitshell(bidon,0));
}

int mv(const char *src, const char *dst){ // déplace fichier avec "mv", renvoie le code de retour de "mv" (0 si OK)
 pid_t pid = fork();
   if (pid == -1) {
      erreur("mv: échec de fork()");
      exit(EXIT_FAILURE);
   }
   else if (pid == 0) {
	   // process fils
	   execlp("mv","mv",src,dst,NULL);
       erreur("exec mv est revenu");
       _exit(EXIT_FAILURE);
   }
   else {
      int status;
      (void)waitpid(pid, &status, 0);
	   return status;
   }
}
